﻿#region License Information

/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */

#endregion

using Microsoft.VisualStudio.TestTools.UnitTesting;
using HeuristicLab.Problems.DataAnalysis;
using System.Globalization;

namespace Tests {
  [TestClass]
  public class ShapeConstraintsParserTest {
    [TestMethod]
    public void TestFunctionRangeConstraints() {
      System.Threading.Thread.CurrentThread.CurrentCulture = CultureInfo.InvariantCulture;
      {
        var constraint = ShapeConstraintsParser.ParseFunctionRangeConstraint("f in [1 .. 2]");
        Assert.AreEqual(string.Empty, constraint.Variable);
        Assert.AreEqual(0, constraint.NumberOfDerivations);
        Assert.AreEqual(false, constraint.IsDerivative);
        Assert.AreEqual(1.0, constraint.Weight);
        Assert.AreEqual(0, constraint.Regions.Count);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
      {
        var constraint = ShapeConstraintsParser.ParseFunctionRangeConstraint("f in [1 2]");
        Assert.AreEqual(string.Empty, constraint.Variable);
        Assert.AreEqual(0, constraint.NumberOfDerivations);
        Assert.AreEqual(false, constraint.IsDerivative);
        Assert.AreEqual(1.0, constraint.Weight);
        Assert.AreEqual(0, constraint.Regions.Count);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
      {
        var constraint = ShapeConstraintsParser.ParseFunctionRangeConstraint("f in [1; 2]");
        Assert.AreEqual(string.Empty, constraint.Variable);
        Assert.AreEqual(0, constraint.NumberOfDerivations);
        Assert.AreEqual(false, constraint.IsDerivative);
        Assert.AreEqual(1.0, constraint.Weight);
        Assert.AreEqual(0, constraint.Regions.Count);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
      {
        var constraint = ShapeConstraintsParser.ParseFunctionRangeConstraint("f in [1 .. 2] weight: 2.0");
        Assert.AreEqual(string.Empty, constraint.Variable);
        Assert.AreEqual(0, constraint.NumberOfDerivations);
        Assert.AreEqual(false, constraint.IsDerivative);
        Assert.AreEqual(2.0, constraint.Weight);
        Assert.AreEqual(0, constraint.Regions.Count);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
      {
        var constraint = ShapeConstraintsParser.ParseFunctionRangeConstraint("f in [1 .. 2], x in [2 .. 3], y in [4 .. 5]");
        Assert.AreEqual(string.Empty, constraint.Variable);
        Assert.AreEqual(0, constraint.NumberOfDerivations);
        Assert.AreEqual(false, constraint.IsDerivative);
        Assert.AreEqual(1.0, constraint.Weight);
        Assert.AreEqual(2, constraint.Regions.Count);
        Assert.AreEqual(2, constraint.Regions.GetInterval("x").LowerBound);
        Assert.AreEqual(3, constraint.Regions.GetInterval("x").UpperBound);
        Assert.AreEqual(4, constraint.Regions.GetInterval("y").LowerBound);
        Assert.AreEqual(5, constraint.Regions.GetInterval("y").UpperBound);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
      {
        var constraint = ShapeConstraintsParser.ParseFunctionRangeConstraint("f in [1 .. 2], x in [2 .. 3], y in [4 .. 5] weight: 2.0");
        Assert.AreEqual(string.Empty, constraint.Variable);
        Assert.AreEqual(0, constraint.NumberOfDerivations);
        Assert.AreEqual(false, constraint.IsDerivative);
        Assert.AreEqual(2.0, constraint.Weight);
        Assert.AreEqual(2, constraint.Regions.Count);
        Assert.AreEqual(2, constraint.Regions.GetInterval("x").LowerBound);
        Assert.AreEqual(3, constraint.Regions.GetInterval("x").UpperBound);
        Assert.AreEqual(4, constraint.Regions.GetInterval("y").LowerBound);
        Assert.AreEqual(5, constraint.Regions.GetInterval("y").UpperBound);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
    }
    [TestMethod]
    public void TestDerivationConstraints() {
      System.Threading.Thread.CurrentThread.CurrentCulture = CultureInfo.InvariantCulture;
      var variables = new string[] { "x", "var x" };
      {
        var constraint = ShapeConstraintsParser.ParseDerivationConstraint("df/dx in [1 .. 2]");
        Assert.AreEqual("x", constraint.Variable);
        Assert.AreEqual(1, constraint.NumberOfDerivations);
        Assert.AreEqual(true, constraint.IsDerivative);
        Assert.AreEqual(1.0, constraint.Weight);
        Assert.AreEqual(0, constraint.Regions.Count);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
      {
        var constraint = ShapeConstraintsParser.ParseDerivationConstraint("d²f/dx² in [1 .. 2]");
        Assert.AreEqual("x", constraint.Variable);
        Assert.AreEqual(2, constraint.NumberOfDerivations);
        Assert.AreEqual(true, constraint.IsDerivative);
        Assert.AreEqual(1.0, constraint.Weight);
        Assert.AreEqual(0, constraint.Regions.Count);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
      {
        var constraint = ShapeConstraintsParser.ParseDerivationConstraint("d² f / d x² in [1 .. 2]");
        Assert.AreEqual("x", constraint.Variable);
        Assert.AreEqual(2, constraint.NumberOfDerivations);
        Assert.AreEqual(true, constraint.IsDerivative);
        Assert.AreEqual(1.0, constraint.Weight);
        Assert.AreEqual(0, constraint.Regions.Count);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
      {
        var constraint = ShapeConstraintsParser.ParseDerivationConstraint("d³ f / d 'var x'³ in [1 .. 2]");
        Assert.AreEqual("var x", constraint.Variable);
        Assert.AreEqual(3, constraint.NumberOfDerivations);
        Assert.AreEqual(true, constraint.IsDerivative);
        Assert.AreEqual(1.0, constraint.Weight);
        Assert.AreEqual(0, constraint.Regions.Count);
        Assert.AreEqual(1.0, constraint.Interval.LowerBound);
        Assert.AreEqual(2.0, constraint.Interval.UpperBound);
      }
    }
  }
}
